"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const grpc = require("grpc");
const Observable_1 = require("rxjs/Observable");
function createServer(protoPath, ns) {
    const server = new grpc.Server();
    const root = grpc.load(protoPath);
    const nsPkg = lookupNamespace(root, ns);
    const serviceNames = getServiceNames(nsPkg);
    serviceNames.forEach(serviceName => {
        server[`add${serviceName}`] = (impl) => {
            const serviceClient = nsPkg[serviceName];
            server.addService(serviceClient.service, createService(nsPkg[serviceName], impl));
        };
    });
    return server;
}
exports.createServer = createServer;
function createClient(protoPath, ns) {
    const root = grpc.load(protoPath);
    const nsPkg = lookupNamespace(root, ns);
    const serviceNames = getServiceNames(nsPkg);
    const client = {};
    serviceNames.forEach(serviceName => {
        client[`get${serviceName}`] = (address, credentials, options) => {
            const clientConstructor = nsPkg[serviceName];
            const serviceClient = new clientConstructor(address, credentials, options);
            return createServiceClient(serviceClient);
        };
    });
    return client;
}
exports.createClient = createClient;
// Inspired, copied and modified from https://github.com/kondi/rxjs-grpc/blob/master/src/index.ts (MIT License)
function lookupNamespace(root, ns) {
    let pkg = root;
    for (const name of ns.split(/\./)) {
        pkg = pkg[name];
    }
    return pkg;
}
function getServiceNames(pkg) {
    return Object.keys(pkg).filter(name => pkg[name].service);
}
function createService(sourceService, rxImpl) {
    const service = {};
    Object.keys(sourceService.prototype).forEach(methodName => {
        if (typeof rxImpl[methodName] === 'function') {
            service[methodName] = createMethod(rxImpl, methodName, sourceService.prototype);
        }
    });
    return service;
}
function createMethod(rxImpl, name, serviceMethods) {
    const methodInfo = serviceMethods[name];
    if (methodInfo.responseStream && methodInfo.requestStream) {
        return createBidirectionalStreamingMethod(rxImpl, name);
    }
    else if (methodInfo.responseStream) {
        return createResponseStreamingMethod(rxImpl, name);
    }
    else if (methodInfo.requestStream) {
        return createRequestStreamingMethod(rxImpl, name);
    }
    else {
        return createUnaryMethod(rxImpl, name);
    }
}
function createUnaryMethod(rxImpl, name) {
    return function (call, callback) {
        const response = rxImpl[name](call.request, call.metadata);
        response.subscribe(data => callback(null, data), error => callback(error));
    };
}
function createResponseStreamingMethod(rxImpl, name) {
    return function (call) {
        return __awaiter(this, void 0, void 0, function* () {
            const response = rxImpl[name](call.request, call.metadata);
            yield response.forEach(data => call.write(data));
            call.end();
        });
    };
}
function createRequestStreamingMethod(rxImpl, name) {
    return function (call, callback) {
        const requestStream = new Observable_1.Observable(subscriber => {
            call.on('data', request => {
                subscriber.next(request);
            });
            call.on('end', () => {
                subscriber.complete();
            });
        });
        const response = rxImpl[name](requestStream, call.metadata);
        response.subscribe(data => callback(null, data), error => callback(error));
    };
}
function createBidirectionalStreamingMethod(rxImpl, name) {
    return function (call) {
        return __awaiter(this, void 0, void 0, function* () {
            const requestStream = new Observable_1.Observable(subscriber => {
                call.on('data', request => {
                    subscriber.next(request);
                });
                call.on('end', () => {
                    subscriber.complete();
                });
            });
            const response = rxImpl[name](requestStream, call.metadata);
            yield response.forEach(data => call.write(data));
            call.end();
        });
    };
}
function createServiceClient(grpcClient) {
    const rxClient = Object.create(grpcClient);
    for (const name of Object.keys(grpcClient.__proto__)) {
        rxClient[name] = createClientMethod(grpcClient, name);
    }
    return rxClient;
}
function createClientMethod(grpcClient, name) {
    const methodInfo = grpcClient[name];
    if (methodInfo.responseStream && methodInfo.requestStream) {
        return createBiDirectionalStreamingClientMethod(grpcClient, name);
    }
    else if (methodInfo.responseStream) {
        return createResponseStreamingClientMethod(grpcClient, name);
    }
    else if (methodInfo.requestStream) {
        return createRequestStreamingClientMethod(grpcClient, name);
    }
    else {
        return createUnaryClientMethod(grpcClient, name);
    }
}
function createUnaryClientMethod(grpcClient, name) {
    return function (request, metaData = new grpc.Metadata(), options = {}) {
        return Observable_1.Observable.create(observer => {
            grpcClient[name](request, metaData, options, (error, data) => {
                if (error) {
                    observer.error(error);
                }
                else {
                    observer.next(data);
                }
                observer.complete();
            });
        });
    };
}
function createResponseStreamingClientMethod(grpcClient, name) {
    return function (request, metaData = new grpc.Metadata(), options = {}) {
        return Observable_1.Observable.create(observer => {
            const call = grpcClient[name](request, metaData, options);
            call.on('data', (data) => observer.next(data));
            call.on('error', (error) => observer.error(error));
            call.on('end', () => observer.complete());
        });
    };
}
function createRequestStreamingClientMethod(grpcClient, name) {
    return function (request, metaData = new grpc.Metadata(), options = {}) {
        return Observable_1.Observable.create(observer => {
            const call = grpcClient[name](metaData, options, (error, data) => {
                if (error) {
                    observer.error(error);
                }
                else {
                    observer.next(data);
                }
                observer.complete();
            });
            request.subscribe(data => call.write(data), error => call.destroy(error), () => call.end());
        });
    };
}
function createBiDirectionalStreamingClientMethod(grpcClient, name) {
    return function (request, metaData = new grpc.Metadata(), options = {}) {
        return Observable_1.Observable.create(observer => {
            const call = grpcClient[name](metaData, options);
            call.on('data', data => observer.next(data));
            call.on('error', error => observer.error(error));
            call.on('end', () => observer.complete());
            request.subscribe(data => call.write(data), error => call.destroy(error), () => call.end());
        });
    };
}
//# sourceMappingURL=GrpcRx.js.map