"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const protobuf = require("protobufjs");
const protobufjs_1 = require("protobufjs");
const fs = require("fs");
const util = require("util");
const NestedWriter_1 = require("./utils/NestedWriter");
function generateTsFromFile(files) {
    return __awaiter(this, void 0, void 0, function* () {
        const root = new protobuf.Root();
        yield root.load(files);
        return generateFromRoot(root);
    });
}
exports.generateTsFromFile = generateTsFromFile;
function generateTsFromString(...protoSources) {
    return __awaiter(this, void 0, void 0, function* () {
        const root = new protobuf.Root();
        protoSources.forEach(source => protobuf.parse(source, root));
        return generateFromRoot(root);
    });
}
exports.generateTsFromString = generateTsFromString;
function generateFromRoot(root) {
    const target = new NestedWriter_1.default();
    target.writeLine(`import { Observable } from 'rxjs/Observable';`);
    target.writeLine(`import * as grpc from 'grpc';`).writeLine();
    print(root.nested, target);
    return target.toString();
}
function print(nested, target) {
    if (!nested) {
        return;
    }
    Object.keys(nested).forEach(ns => {
        const o = nested[ns];
        if (o instanceof protobufjs_1.Service) {
            printService(o, target);
        }
        else if (o instanceof protobufjs_1.Enum) {
            printEnum(o, target);
        }
        else if (o instanceof protobufjs_1.Type) {
            printType(o, target);
        }
        else if (o instanceof protobufjs_1.Namespace) {
            printNamespace(o, target);
        }
    });
}
function printNamespace(source, target) {
    if (source.comment) {
        printComment(source.comment, target);
    }
    target.writeLine('export namespace ' + source.name + ' {').inc();
    print(source.nested, target);
    printServer(source.nested, target);
    printClient(source.nested, target);
    target.dec().writeLine('}').writeLine();
}
function printServer(nested, target) {
    if (!nested) {
        return;
    }
    const services = Object.keys(nested)
        .map(ns => nested[ns])
        .filter(o => o instanceof protobufjs_1.Service);
    if (!services.length) {
        return;
    }
    target.writeLine('export interface Server extends grpc.Server {').inc();
    services.forEach((service) => {
        target.writeLine(`add${service.name}(impl: ${service.name});`);
    });
    target.dec().writeLine('}').writeLine();
}
function printClient(nested, target) {
    if (!nested) {
        return;
    }
    const services = Object.keys(nested)
        .map(ns => nested[ns])
        .filter(o => o instanceof protobufjs_1.Service);
    if (!services.length) {
        return;
    }
    services.forEach((service) => {
        target.writeLine(`export interface ${service.name}Client extends grpc.Client, ${service.name} {}`);
    });
    target.writeLine();
    target.writeLine('export interface ClientBuilder {').inc();
    services.forEach((service) => {
        target.writeLine(`get${service.name}(address: string, credentials: grpc.ChannelCredentials, options?: object): ${service.name}Client;`);
    });
    target.dec().writeLine('}').writeLine();
}
function printComment(comment, target) {
    target.writeLine();
    target.writeLine('/**');
    comment.split('\n').forEach(line => {
        target.writeLine(' * ' + line);
    });
    target.writeLine(' */');
}
function printService(source, target) {
    if (source.comment) {
        printComment(source.comment, target);
    }
    target.writeLine('export interface ' + source.name + ' {').inc();
    for (let method of Object.keys(source.methods)) {
        printMethod(source.methods[method], target);
    }
    target.dec().writeLine('}').writeLine();
}
function printMethod(source, target) {
    if (source.comment) {
        printComment(source.comment, target);
    }
    const reqType = source.requestStream ? `Observable<${source.requestType}>` : source.requestType;
    target.writeLine(`${source.name}(request: ${reqType}, metaData?: grpc.Metadata): Observable<${source.responseType}>;`);
}
function printType(source, target) {
    if (source.comment) {
        printComment(source.comment, target);
    }
    target.writeLine('export interface ' + source.name + ' {').inc();
    for (let field of Object.keys(source.fields)) {
        printField(source.fields[field], target);
    }
    target.dec().writeLine('}').writeLine();
}
function printEnum(source, target) {
    if (source.comment) {
        printComment(source.comment, target);
    }
    target.writeLine('export enum ' + source.name + ' {').inc();
    for (let value of Object.keys(source.values)) {
        target.writeLine(`${value} = ${source.values[value]},`);
    }
    target.dec().writeLine('}').writeLine();
}
function printField(source, target) {
    if (source.comment) {
        printComment(source.comment, target);
    }
    target.writeLine(source.name + ': ' + source.type + ';');
}
function cli() {
    let output = 'stdout';
    const files = process.argv.slice(2);
    for (let i = 0; i < files.length; i++) {
        const arg = files[i];
        if (arg === '-o') {
            output = files[i + 1];
            files.splice(i, 2);
            i--;
        }
    }
    if (!files.length) {
        console.log('No files');
        console.log('Usage:');
        console.log('grpc-node-rxjs file1.proto [file2.proto] [-o namespaces.ts]');
        process.exit(1);
    }
    generateTsFromFile(files).then(result => {
        if (output === 'stdout') {
            process.stdout.write(result);
        }
        else {
            return util.promisify(fs.writeFile)(output, result);
        }
    }).catch(console.error);
}
exports.cli = cli;
if (!module.parent) {
    cli();
}
//# sourceMappingURL=generateTs.js.map